@@ -361,9 +361,9 @@ RADEONGetBIOSInfo(ScrnInfoPtr pScrn, xf86Int10InfoPtr  pInt10)
 
 #ifdef XSERVER_LIBPCIACCESS
     int size = info->PciInfo->rom_size > RADEON_VBIOS_SIZE ? info->PciInfo->rom_size : RADEON_VBIOS_SIZE;
-    info->VBIOS = xalloc(size);
+    info->VBIOS = malloc(size);
 #else
-    info->VBIOS = xalloc(RADEON_VBIOS_SIZE);
+    info->VBIOS = malloc(RADEON_VBIOS_SIZE);
 #endif
     if (!info->VBIOS) {
 	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
@@ -381,7 +381,7 @@ RADEONGetBIOSInfo(ScrnInfoPtr pScrn, xf86Int10InfoPtr  pInt10)
     if (info->VBIOS[0] != 0x55 || info->VBIOS[1] != 0xaa) {
 	xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
 		   "Unrecognized BIOS signature, BIOS data will not be used\n");
-	xfree (info->VBIOS);
+	free (info->VBIOS);
 	info->VBIOS = NULL;
 	return FALSE;
     }
@@ -396,7 +396,7 @@ RADEONGetBIOSInfo(ScrnInfoPtr pScrn, xf86Int10InfoPtr  pInt10)
     else if (info->VBIOS[dptr + 0x14] != 0x0) {
 	xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
 		   "Not an x86 BIOS ROM image, BIOS data will not be used\n");
-	xfree (info->VBIOS);
+	free (info->VBIOS);
 	info->VBIOS = NULL;
 	return FALSE;
     }
@@ -406,7 +406,7 @@ RADEONGetBIOSInfo(ScrnInfoPtr pScrn, xf86Int10InfoPtr  pInt10)
     if(!info->ROMHeaderStart) {
 	xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
 		   "Invalid ROM pointer, BIOS data will not be used\n");
-	xfree (info->VBIOS);
+	free (info->VBIOS);
 	info->VBIOS = NULL;
 	return FALSE;
     }
@@ -548,15 +548,6 @@ static void RADEONApplyLegacyQuirks(ScrnInfoPtr pScrn, int index)
 	}
     }
 
-    /* Some RV100 cards with 2 VGA ports show up with DVI+VGA */
-    if (info->Chipset == PCI_CHIP_RV100_QY &&
-	PCI_SUB_VENDOR_ID(info->PciInfo) == 0x1002 &&
-	PCI_SUB_DEVICE_ID(info->PciInfo) == 0x013a) {
-	if (info->BiosConnector[index].ConnectorType == CONNECTOR_DVI_I) {
-	    info->BiosConnector[index].ConnectorType = CONNECTOR_VGA;
-	}
-    }
-
     /* X300 card with extra non-existent DVI port */
     if (info->Chipset == PCI_CHIP_RV370_5B60 &&
 	PCI_SUB_VENDOR_ID(info->PciInfo) == 0x17af &&
